from .base import Question
from prompt_toolkit.layout.controls import FormattedTextControl, BufferControl
from prompt_toolkit.layout import HSplit, VSplit, Window
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.filters.app import has_focus
from prompt_toolkit.document import Document


class ChoiceCandidatesControl(FormattedTextControl):
    def __init__(self, candidates, default=0):
        self._candidates = candidates
        self.selected_item = default
        super().__init__(self._build_text, show_cursor=False, focusable=True)

    def _build_text(self):
        tokens = []
        for i, c in enumerate(self._candidates):
            if i == self.selected_item:
                tokens.extend(
                    [("class:choice.cursor", "❯ "), ("class:choice.selected", c)]
                )
            else:
                tokens.extend([("", "  "), ("class:choice.unselected", c)])
            tokens.append(("", "\n"))

        tokens.pop()
        return tokens


class ChoiceOtherLabelControl(FormattedTextControl):
    def __init__(self, selected=False, label=None):
        self.selected = selected
        self.label = "other" if label is None else label
        super().__init__(self._build_text, show_cursor=False, focusable=False)

    def _build_text(self):
        if self.selected:
            return [
                ("class:choice.cursor", "❯ "),
                ("class:choice.selected", self.label),
                ("class:choice.selected", ": "),
            ]
        else:
            return [
                ("", "  "),
                ("class:choice.unselected", self.label),
                ("class:choice.unselected", ": "),
            ]


class ChoiceQuestion(Question):
    def __init__(self, question, candidates, other=False, default=None, **kwargs):
        self._candidates = candidates

        self._other = other
        self._selected, other_value = self._get_default_selected(default)

        self._candidate_control = ChoiceCandidatesControl(candidates, self._selected)
        if self.has_other:
            buf = Buffer(accept_handler=self._buffer_handler, multiline=False, **kwargs)
            buf.document = Document(other_value)
            self._other_value_control = BufferControl(buf)
            self._other_label_control = ChoiceOtherLabelControl(
                self.other_selected,
                self._other if isinstance(self._other, str) else None,
            )

        super().__init__(question)

    def build_question(self):
        w = [Window(self._candidate_control, dont_extend_height=True)]
        if self.has_other:
            w.append(
                VSplit(
                    [
                        Window(
                            self._other_label_control,
                            dont_extend_width=True,
                            dont_extend_height=True,
                        ),
                        Window(self._other_value_control, dont_extend_height=True),
                    ]
                )
            )

        return HSplit(w, key_bindings=self._get_key_bindings())

    def _get_key_bindings(self):
        kb = KeyBindings()

        on_sel = ~has_focus(self._other_value_control) if self.has_other else True

        @kb.add("j", filter=on_sel)
        @kb.add("down", filter=on_sel)
        @kb.add("c-n", filter=on_sel)
        def _(event):
            self.move_cursor(1)

        @kb.add("k", filter=on_sel)
        @kb.add("up", filter=on_sel)
        @kb.add("c-p", filter=on_sel)
        def _(event):
            self.move_cursor(-1)

        @kb.add("enter", filter=on_sel)
        def _(event):
            if self.other_selected:
                return event.app.layout.focus(self._other_value_control)

            self.set_answer(self._candidates[self._selected])

        return kb

    def _buffer_handler(self, doc):
        self.set_answer(doc.text)

    def _get_default_selected(self, default):
        if default is None:
            return 0, ""

        if not self.has_other:
            return self._candidates.index(default), ""

        try:
            return self._candidates.index(default), ""
        except ValueError:
            return len(self._candidates), default

    @property
    def other_selected(self):
        return self._selected == len(self._candidates)

    @property
    def has_other(self):
        return self._other is not None and self._other is not False

    @property
    def num_candidates(self):
        return len(self._candidates) + (1 if self.has_other else 0)

    def move_cursor(self, d):
        self._selected = (self._selected + d) % self.num_candidates
        if self.has_other:
            if self.other_selected:
                self._candidate_control.selected_item = None
                self._other_label_control.selected = True
            else:
                self._candidate_control.selected_item = self._selected
                self._other_label_control.selected = False
        else:
            self._candidate_control.selected_item = self._selected

    def main_control(self):
        return self._candidate_control



from .wizard import wizard

from .question.text import TextQuestion
from .question.choice import ChoiceQuestion
from .question.yesno import YesNoQuestion

__all__ = ("wizard", "TextQuestion", "ChoiceQuestion", "YesNoQuestion")



from prompt_toolkit.layout import Window

from .base import BaseWidget, WrapFormattedTextControl


class Message(BaseWidget):
    def __init__(self, message):
        self._widget = Window(
            WrapFormattedTextControl(message),
            dont_extend_height=True,
            dont_extend_width=True,
        )

    def __pt_container__(self):
        return self._widget



from .base import Question
from prompt_toolkit.key_binding import KeyBindings


def defalt_pretty(v):
    return "Yes" if v else "No"


def default_converter(v):
    return True if v.lower() == "y" else False


class YesNoQuestion(Question):
    def __init__(self, question, keys="Yn", pretty=None, converter=None):
        self._keys = keys
        self._pretty = defalt_pretty if pretty is None else pretty
        self._converter = default_converter if converter is None else converter
        self._question_text = ("class:question.text", question)
        self._question_key = ("class::question.key", " [{}]".format(keys))
        super().__init__([self._question_text, self._question_key])

    def key_bindings(self):
        kb = KeyBindings()

        default = None

        for key in self._keys:
            if key.isupper():
                default = key

            @kb.add(key.upper(), eager=True)
            @kb.add(key.lower(), eager=True)
            def _(event, key=key):
                self.question = [self._question_text]
                self.set_answer(self._converter(key), pretty=self._pretty)

        if default is not None:

            @kb.add("enter", eager=True)
            def _(event):
                self.question = [self._question_text]
                self.set_answer(self._converter(default), pretty=self._pretty)

        return kb



from abc import ABCMeta, abstractmethod
from prompt_toolkit.layout import ConditionalContainer, Window
from prompt_toolkit.layout.containers import HSplit, VSplit
from prompt_toolkit.filters import Condition
from prompt_toolkit.widgets.toolbars import ValidationToolbar
from prompt_toolkit.layout.controls import FormattedTextControl


class BaseWidget(metaclass=ABCMeta):
    @abstractmethod
    def __pt_container__(self):
        raise NotImplementedError

    def main_control(self):
        return None


class WrapFormattedTextControl(FormattedTextControl):
    def preferred_height(
        self, width, max_available_height, wrap_lines, get_line_prefix
    ):
        content = self.create_content(width, None)
        total = 0
        for i in range(content.line_count):
            total += content.get_height_for_line(i, width, get_line_prefix)

        return total


class QuestionHeaderControl(WrapFormattedTextControl):
    def __init__(self, question):
        self.question = question
        super().__init__(self._build_text, show_cursor=False, focusable=True)

    def _build_text(self):
        tokens = [("class:question.mark", "?"), ("", " ")]
        if isinstance(self.question, list):
            tokens.extend(self.question)
        else:
            tokens.append(("class:question.text", self.question))

        return tokens


class AnswerControl(WrapFormattedTextControl):
    def __init__(self, answer=None):
        self.answer = answer
        super().__init__(self._build_text, show_cursor=False, focusable=False)

    def _build_text(self):
        return [("", " "), ("class:answer.text", self.answer or "")]


class Question(BaseWidget):
    def __init__(self, question):
        self._answer_control = AnswerControl()

        @Condition
        def answered():
            return self.answer is not None

        self.answered = answered

        q = self.build_question()

        self._header_control = QuestionHeaderControl(question)
        widgets = [
            VSplit(
                [
                    Window(
                        self._header_control,
                        wrap_lines=True,
                        dont_extend_height=True,
                        dont_extend_width=True,
                    ),
                    ConditionalContainer(
                        Window(
                            self._answer_control,
                            wrap_lines=True,
                            dont_extend_height=True,
                        ),
                        self.answered,
                    ),
                ]
            )
        ]
        if q is not None:
            widgets.append(ConditionalContainer(q, ~self.answered))

        widgets.extend([ConditionalContainer(ValidationToolbar(), ~self.answered)])

        self._widget = HSplit(widgets, key_bindings=self.key_bindings())

    def build_question(self):
        return

    def __pt_container__(self):
        return self._widget

    @property
    def answer(self):
        return self._answer_control.answer

    def set_answer(self, value, pretty=None):
        if value is None:
            self._answer_control.answer = None
            return

        if pretty is None:
            pretty = default_pretty

        self._answer_control.answer = pretty(value)
        self._handler(value)

    @property
    def question(self):
        return self._header_control.question

    @question.setter
    def question(self, value):
        self._header_control.question = value

    def key_bindings(self):
        return None

    def main_control(self):
        return self._header_control


def default_pretty(v, conv=str):
    return v if isinstance(v, conv) else str(v)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['aleister', 'aleister.question']

package_data = \
{'': ['*']}

install_requires = \
['prompt_toolkit>=2.0,<3.0']

setup_kwargs = {
    'name': 'aleister',
    'version': '0.1.1',
    'description': 'library for CUI wizard',
    'long_description': None,
    'author': 'Hirotomo Moriwaki',
    'author_email': 'philopon.dependence@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



from prompt_toolkit import Application
from prompt_toolkit.application import get_app
from prompt_toolkit.layout.layout import Layout, Window
from prompt_toolkit.layout.controls import DummyControl
from prompt_toolkit.layout.containers import HSplit
from prompt_toolkit.styles import Style
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.patch_stdout import patch_stdout
from functools import wraps


class Cancel(Exception):
    pass


class Wizard:
    def __init__(self, questions, key_bindings=None, style=None):
        self._questions_iter = iter(questions)
        self._key_bindings = (
            _get_default_key_bindings() if key_bindings is None else key_bindings
        )
        self._style = _get_default_style() if style is None else style

    def run(self):
        answer = None
        while True:
            try:
                widget = self._questions_iter.send(answer)
            except StopIteration:
                break

            layout = Layout(widget)
            app = Application(
                layout,
                full_screen=False,
                key_bindings=self._key_bindings,
                style=self._style,
            )
            c = widget.main_control()
            if c is not None:
                app.layout.focus(c)
            widget._handler = self._make_handler(app)

            try:
                answer = app.run()
            except Cancel:
                break

    def _make_handler(self, app):
        def handler(answer):
            app.exit(answer)

        return handler


def _get_default_style():
    return Style.from_dict(
        {
            "question.mark": "ansiblue bold",
            "question.text": "bold",
            "question.key": "",
            "answer.text": "ansiyellow bold",
            "choice.selected": "bold",
            "choice.unselected": "",
            "choice.cursor": "ansired bold",
        }
    )


def _get_default_key_bindings():
    k = KeyBindings()

    @k.add("c-c")
    def _(event):
        event.app.exit(exception=Cancel())

    return k


def wizard(style=None, key_bindings=None):
    def wrapped(f):
        @wraps(f)
        def g(*args, **kwargs):
            wiz = Wizard(f(*args, *kwargs), key_bindings=key_bindings, style=style)
            wiz.run()

        return g

    return wrapped



from .base import Question
from prompt_toolkit.layout.containers import Window
from prompt_toolkit.layout.controls import BufferControl
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.document import Document


class TextQuestion(Question):
    def __init__(self, question, default=None, **kwargs):
        self._kwargs = kwargs
        buf = Buffer(multiline=False, accept_handler=self._accept, **self._kwargs)
        self._buf_control = BufferControl(buf)
        if default is not None:
            buf.document = Document(default)

        super().__init__(question)

    def main_control(self):
        return self._buf_control

    def build_question(self):
        return Window(self._buf_control, dont_extend_height=True, wrap_lines=True)

    def _accept(self, doc):
        self.set_answer(doc.text)
